from __future__ import annotations

from matic.utils.base_token import BaseToken
from matic.utils.web3_side_chain_client import Web3SideChainClient


class RootChain(BaseToken):
    """Root chain implementation.

    This represents a connection between parent (root) and child chains.
    For example, Goerli testnet is a root chain for Mumbai testnet.
    """

    def __init__(self, client: Web3SideChainClient, address: str):
        super().__init__(
            address=address,
            name='RootChain',
            is_parent=True,
            client=client,
        )

    @property
    def last_child_block(self) -> int:
        """Get last block number on child chain."""
        return self.method('getLastChildBlock').read()

    def find_root_block_from_child(self, child_block_number: int) -> int:
        """Find root block corresponding to child block of given number."""
        checkpoint_interval = 10000

        # First checkpoint id = start * 10000
        start = 1
        # Last checkpoint id = end * 10000
        method = self.method('currentHeaderBlock')
        current_header_block = method.read()
        end = int(current_header_block) // checkpoint_interval

        # Binary search on all the checkpoints to find the checkpoint
        # that contains the child_block_number
        ans = None
        while start <= end:
            if start == end:
                ans = start
                break
            mid = (start + end) // 2
            _, header_start, header_end, _, _ = self.method(
                'headerBlocks', mid * checkpoint_interval
            ).read()

            if header_start <= child_block_number <= header_end:
                # If child_block_number is between the upper and lower bounds
                # of the header_block, we found our answer
                ans = mid
                break
            elif header_start > child_block_number:
                # // child_block_number was checkpointed before self header
                end = mid - 1
            elif header_end < child_block_number:
                # // child_block_number was checkpointed after self header
                start = mid + 1

        assert ans is not None
        return ans * checkpoint_interval



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Sequence

from typing_extensions import NotRequired, Required, TypedDict
from web3.types import RPCEndpoint

if TYPE_CHECKING:
    from matic.pos.exit_util import ExitUtil
    from matic.pos.root_chain_manager import RootChainManager


ConfigWithFrom = TypedDict('ConfigWithFrom', {'from': Required[str]})
"""Configuration dictionary with required key "from" (type str) and any other keys."""


class ITransactionRequestConfig(ConfigWithFrom, total=False):
    """Transaction config - an actual configuration used to interact with chain."""

    to: str
    value: int
    gas_limit: int
    gas_price: int
    data: bytes
    nonce: int
    chain_id: int
    chain: str
    hardfork: str
    max_fee_per_gas: int
    max_priority_fee_per_gas: int
    type: int


class ITransactionOption(ITransactionRequestConfig):
    """Transaction config: this can be passed as option to almost all methods."""

    return_transaction: NotRequired[bool]


class IAllowanceTransactionOption(ITransactionOption):
    """Transaction config for approve/allowance methods."""

    spender_address: NotRequired[bytes]
    """Address of spender.

    **spender** - third-party user or a smart contract which can transfer
    your token on your behalf.
    """


class IExitTransactionOption(ITransactionOption):
    """Transaction config for ``withdraw_exit_*`` operations."""

    burn_event_signature: NotRequired[bytes]


@dataclass
class ITransactionData:
    """Transaction parameters that can be obtained from blockchain."""

    transaction_hash: bytes
    nonce: int
    block_hash: bytes | None
    block_number: int | None
    transaction_index: int | None
    from_: str
    to: str | None
    value: int
    gas_price: int
    gas: int
    input: bytes


class NeighbourClientConfig(TypedDict):
    """Configuration for parent/child of :class:`~matic.utils.web3_side_chain_client.Web3SideChainClient`."""  # noqa

    provider: Any
    default_config: ConfigWithFrom


class IBaseClientConfig(TypedDict):
    """Configuration for :class:`~matic.utils.web3_side_chain_client.Web3SideChainClient`."""  # noqa

    network: str
    """Network to connect - 'testnet' or 'mainnet'."""
    version: str
    """Network version - 'mumbai' or 'v1'."""
    parent: NotRequired[NeighbourClientConfig]
    """Parent chain configuration."""
    child: NotRequired[NeighbourClientConfig]
    """Child chain configuration."""
    log: NotRequired[bool]


class IPOSClientConfig(IBaseClientConfig):
    """Configuration for POS client."""

    root_chain_manager: NotRequired[str]
    root_chain: NotRequired[str]
    erc_1155_mintable_predicate: NotRequired[str]


@dataclass
class IBaseBlock:
    """Base block parameters."""

    size: int
    difficulty: int
    total_difficulty: int
    uncles: Sequence[bytes]
    number: int
    hash: bytes
    parent_hash: bytes
    nonce: int
    sha3_uncles: bytes
    logs_bloom: bytes
    transactions_root: bytes
    state_root: bytes
    receipts_root: bytes
    miner: bytes
    extra_data: bytes
    gas_limit: int
    gas_used: int
    timestamp: int
    base_fee_per_gas: int | None = None


@dataclass
class IBlock(IBaseBlock):
    """Block with raw transactions (bytes)."""

    transactions: Sequence[bytes] = field(default_factory=list)


@dataclass
class IBlockWithTransaction(IBaseBlock):
    """Block with decoded transactions."""

    transactions: Sequence[ITransactionData] = field(default_factory=list)


@dataclass
class IPOSContracts:
    """Return type of ``get_pos_contracts`` parameter of :class:`matic.pos.POSToken`."""

    root_chain_manager: RootChainManager
    exit_util: ExitUtil


@dataclass
class IRootBlockInfo:
    """Root block info (used in proofs)."""

    start: int
    """Block start number."""
    end: int
    """Block end number."""
    header_block_number: int
    """Header block number - root block number in which child block exist."""


class IJsonRpcRequestPayload(TypedDict):
    """JSON-RPC request parameters dict."""

    method: RPCEndpoint | str
    params: list[Any]
    id: NotRequired[str | int]


class IJsonRpcResponse(TypedDict):
    """JSON-RPC response parameters dict."""

    jsonrpc: str
    id: int
    result: NotRequired[Any]
    error: NotRequired[str]


class ITransactionWriteResult(ABC):
    """Interface for result of ``process_write`` method."""

    @property
    @abstractmethod
    def transaction_hash(self) -> bytes:
        """Get hash of executed transaction."""
        ...

    @property
    def receipt(self) -> ITransactionReceipt:
        """Property for convenient access to receipt."""
        return self.get_receipt()

    @abstractmethod
    def get_receipt(self, timeout: int = ...) -> ITransactionReceipt:
        """Get receipt (wait max ``timeout`` seconds)."""
        ...


@dataclass
class ILog:
    """Log data."""

    address: str
    data: str
    topics: Sequence[bytes]
    log_index: int
    transaction_hash: bytes
    transaction_index: int
    block_hash: bytes
    block_number: int
    removed: bool


class _RawLogData(TypedDict):
    data: bytes
    topics: Sequence[bytes]


@dataclass
class IEventLog:
    """Event logs (can occur in transaction receipts)."""

    event: str
    address: bytes
    return_values: Any
    log_index: int
    transaction_index: int
    transaction_hash: bytes
    block_hash: bytes
    block_number: int
    raw: _RawLogData | None = None


@dataclass
class ITransactionReceipt:
    """Transaction receipt format used internally."""

    transaction_hash: bytes
    transaction_index: int
    block_hash: bytes
    block_number: int
    from_: bytes
    to: bytes
    contract_address: bytes
    cumulative_gas_used: int
    gas_used: int
    logs_bloom: bytes
    root: bytes
    type: str
    status: bool | None = None
    logs: Sequence[ILog] = field(default_factory=list)
    events: dict[str, IEventLog] = field(default_factory=dict)


@dataclass
class CheckpointedBlock(IRootBlockInfo):
    """Block info obtained from API to construct a proof."""

    header_block_number: int
    block_number: int
    start: int
    end: int
    proposer: str  # hex string, with "0x"
    root: str  # hex string, with "0x"
    created_at: int
    message: str



from __future__ import annotations

from typing import Any

import requests

from matic.json_types import CheckpointedBlock
from matic.utils.polyfill import removeprefix, removesuffix

DEFAULT_ABI_STORE_URL: str = 'https://static.matic.network/network'
"""Default url of ABI store. Can be altered if needed."""

DEFAULT_PROOF_API_URL: str = ''
"""Default url of proof API. Must be set to use fast proofs."""


def get_abi(
    network: str,
    version: str,
    bridge_type: str,
    contract_name: str,
    base_url: str | None = None,
) -> dict[str, Any]:
    """Get ABI dict for contract."""
    base_url = removesuffix(base_url or DEFAULT_ABI_STORE_URL, '/')
    url = f'{base_url}/{network}/{version}/artifacts/{bridge_type}/{contract_name}.json'
    return requests.get(url).json()['abi']


def get_address(
    network: str, version: str, base_url: str | None = None
) -> dict[str, Any]:
    """Fetch dictionary with addresses of contracts deployed on network."""
    base_url = removesuffix(base_url or DEFAULT_ABI_STORE_URL, '/')
    url = f'{base_url}/{network}/{version}/index.json'
    return requests.get(url).json()


def _create_proof_url(network: str, url: str, base_url: str | None = None) -> str:
    base_url = removesuffix(base_url or DEFAULT_PROOF_API_URL, '/')
    if not base_url:
        raise RuntimeError('Please set DEFAULT_PROOF_API_URL or pass base_url.')
    url = removeprefix(url, '/')
    return f'{base_url}/{"matic" if network == "mainnet" else "mumbai"}/{url}'


def get_block_included(
    network: str, block_number: int, base_url: str | None = None
) -> CheckpointedBlock:
    """Get block information by number."""
    url = _create_proof_url(network, f'/block-included/{block_number}')
    data = requests.get(url).json()
    return CheckpointedBlock(
        header_block_number=int(data['headerBlockNumber'], 16),
        block_number=int(data['blockNumber']),
        start=int(data['start']),
        end=int(data['end']),
        proposer=data['proposer'],
        root=data['root'],
        created_at=int(data['createdAt']),
        message=data['message'],
    )


def get_proof(
    network: str, start: int, end: int, block_number: int, base_url: str | None = None
) -> bytes:
    """Get proof from API."""
    url = _create_proof_url(
        network,
        f'/fast-merkle-proof?start={start}&end={end}&number={block_number}',
    )
    return bytes.fromhex(removeprefix(requests.get(url).json()['proof'], '0x'))



from __future__ import annotations

from matic.json_types import IPOSClientConfig, IPOSContracts, ITransactionOption
from matic.pos.erc_20 import ERC20
from matic.pos.erc_721 import ERC721
from matic.pos.erc_1155 import ERC1155
from matic.pos.exit_util import ExitUtil
from matic.pos.root_chain import RootChain
from matic.pos.root_chain_manager import RootChainManager
from matic.utils.bridge_client import BridgeClient


class POSClient(BridgeClient):
    """POS bridge client.

    Used to manage instantiation of ``ERC20``, ``ERC721`` and ``ERC1155`` classes
    and perform some common operations.
    """

    root_chain_manager: RootChainManager
    config: IPOSClientConfig

    def __init__(self, config: IPOSClientConfig):
        super().__init__(config)
        main_pos_contracts = self.client.main_pos_contracts
        config['root_chain_manager'] = (
            config.get('root_chain_manager')
            or main_pos_contracts['RootChainManagerProxy']
        )
        config['root_chain'] = (
            config.get('root_chain')
            or self.client.main_plasma_contracts['RootChainProxy']
        )
        self.client.config = config

        self.root_chain_manager = RootChainManager(
            self.client, config['root_chain_manager']
        )

        self.exit_util = ExitUtil(
            self.client, RootChain(self.client, config['root_chain'])
        )

    def erc_20(self, token_address: str, is_parent: bool = False) -> ERC20:
        """Instantiate :class:`~matic.pos.ERC20` for token address.

        Args:
            token_address: address where token contract is deployed.
            is_parent: Whether this belongs to parent or child chain.
        """
        return ERC20(token_address, is_parent, self.client, self._get_contracts)

    def erc_721(self, token_address: str, is_parent: bool = False) -> ERC721:
        """Instantiate :class:`~matic.pos.ERC721` for token address.

        Args:
            token_address: address where token contract is deployed.
            is_parent: Whether this belongs to parent or child chain.
        """
        return ERC721(token_address, is_parent, self.client, self._get_contracts)

    def erc_1155(self, token_address: str, is_parent: bool = False) -> ERC1155:
        """Instantiate :class:`~matic.pos.ERC1155` for token address.

        Args:
            token_address: address where token contract is deployed.
            is_parent: Whether this belongs to parent or child chain.
        """
        return ERC1155(token_address, is_parent, self.client, self._get_contracts)

    def deposit_ether(
        self,
        amount: int,
        user_address: str,
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Deposit given amount of ether to polygon chain."""
        return ERC20(
            '',
            True,
            self.client,
            self._get_contracts,
        )._deposit_ether(amount, user_address, private_key, option)

    def _get_contracts(self) -> IPOSContracts:
        return IPOSContracts(
            exit_util=self.exit_util,
            root_chain_manager=self.root_chain_manager,
        )



from __future__ import annotations

import sys

if sys.version_info < (3, 9):

    def removeprefix(s: str, prefix: str, /) -> str:
        """str.removeprefix ported to old python."""
        if s.startswith(prefix):
            return s[len(prefix) :]
        return s

    def removesuffix(s: str, suffix: str, /) -> str:
        """str.removesuffix ported to old python."""
        if s.endswith(suffix):
            return s[: -len(suffix)]
        return s

else:
    removeprefix = str.removeprefix
    removesuffix = str.removesuffix



from __future__ import annotations

from typing import Any, cast

from matic.abstracts import BaseContract, BaseContractMethod, BaseWeb3Client
from matic.exceptions import (
    AllowedOnChildException,
    AllowedOnRootException,
    EIP1559NotSupportedException,
)
from matic.json_types import (
    ITransactionOption,
    ITransactionRequestConfig,
    ITransactionWriteResult,
)
from matic.utils.web3_side_chain_client import Web3SideChainClient


class BaseToken:
    """Base class for all tokens."""

    _contract: BaseContract | None = None

    def __init__(
        self,
        address: str,
        is_parent: bool,
        name: str,
        client: Web3SideChainClient,
        bridge_type: str | None = None,
    ):
        self.address = address
        self.is_parent = is_parent
        self.name = name
        self.bridge_type = bridge_type
        self.client = client

    @property
    def contract(self) -> BaseContract:
        """Contract object."""
        if self._contract:
            return self._contract

        abi = self.client.get_abi(
            self.name,
            self.bridge_type,
        )
        self._contract = self._get_contract(
            self.is_parent,
            self.address,
            abi,
        )
        assert self._contract
        return self._contract

    def method(self, method_name: str, *args: Any) -> Any:
        """Call arbitrary JSON-RPC method."""
        return self.contract.method(method_name, *args)

    def process_write(
        self,
        method: BaseContractMethod,
        option: ITransactionOption | None = None,
        private_key: str | None = None,
    ) -> ITransactionWriteResult | ITransactionRequestConfig:
        """Perform write (modifying) operation.

        Args:
            method: Method instance (with arguments passed on instantiation)
            option: Additional parameters.
                May contain special key ``return_transaction`` (see Returns section)
            private_key: Sender private key
                (may be missing, if your provider supports implicit tx signing)

        Returns:
            ITransactionWriteResult if actual write was performed;
                ITransactionRequestConfig if `return_transaction=True`.
                (builds the final transaction dictionary and returns it).
        """
        config = self.create_transaction_config(
            tx_config=option,
            is_write=True,
            method=method,
            is_parent=self.is_parent,
        )

        self.client.logger.info('process write config: %s', config)
        if option and option.get('return_transaction', False):
            config['data'] = method.encode_abi()
            config['to'] = method.address
            return config

        return method.write(config, private_key)

    # def send_transaction(
    #     self, option: ITransactionOption | None = None
    # ) -> ITransactionWriteResult | ITransactionRequestConfig:
    #     is_parent = self.is_parent
    #     client = self.get_client(is_parent)

    #     config = self.create_transaction_config(
    #         tx_config=option,
    #         is_write=True,
    #         method=None,
    #         is_parent=self.is_parent,
    #     )

    #     client.logger.info('process write config: %s', config)
    #     if option and option.get('return_transaction', False):
    #         return config

    #     return client.write(config)

    def read_transaction(
        self, option: ITransactionOption | None = None
    ) -> Any | ITransactionRequestConfig:
        """Send read (non-modifying) transaction without RPC method.

        Args:
            option: Additional parameters.
                May contain special key ``return_transaction`` (see Returns section)

        Returns:
            ITransactionWriteResult if actual write was performed;
                ITransactionRequestConfig if `return_transaction=True`.
                (builds the final transaction dictionary and returns it).
        """
        is_parent = self.is_parent
        client = self.get_client(is_parent)

        config = self.create_transaction_config(
            tx_config=option,
            is_write=True,
            method=None,
            is_parent=self.is_parent,
        )

        client.logger.info('process read config: %s', config)
        if option and option.get('return_transaction', False):
            return config

        return client.read(config)

    def process_read(
        self, method: BaseContractMethod, option: ITransactionOption | None = None
    ) -> Any:
        """Perform read (non-modifying) operation.

        Args:
            method: Method instance (with arguments passed on instantiation)
            option: Additional parameters.
                May contain special key ``return_transaction`` (see Returns section)

        Returns:
            ITransactionWriteResult if actual write was performed;
                ITransactionRequestConfig if `return_transaction=True`.
                (builds the final transaction dictionary and returns it).
        """
        config = self.create_transaction_config(
            tx_config=option,
            is_write=False,
            method=method,
            is_parent=self.is_parent,
        )
        self.client.logger.info('read tx config created: %s', config)

        if option and option.get('return_transaction', False):
            assert self.contract
            config.update({'data': method.encode_abi(), 'to': self.contract.address})
            return config

        return method.read(config)

    def get_client(self, is_parent: bool) -> BaseWeb3Client:
        """Get web3 client instance."""
        if is_parent:
            return self.client.parent
        return self.client.child

    def _get_contract(
        self, is_parent: bool, token_address: str, abi: dict[str, Any]
    ) -> BaseContract:
        client = self.get_client(is_parent)
        return client.get_contract(token_address, abi)

    def create_transaction_config(
        self,
        tx_config: ITransactionRequestConfig | None,
        method: BaseContractMethod | None,
        is_parent: bool,
        is_write: bool,
    ) -> ITransactionRequestConfig:
        """Fill in missing fields in transaction request.

        Warning: this method may raise if your transaction cannot be executed,
            pass ``gas_limit`` to prevent it from happening.
        """
        if is_parent:
            default_config = (self.client.config.get('parent') or {}).get(
                'default_config'  # type: ignore
            )
        else:
            default_config = (self.client.config.get('child') or {}).get(
                'default_config'  # type: ignore
            )

        merged_config = dict(default_config or {})
        merged_config.update(tx_config or {})

        tx_config = cast(ITransactionRequestConfig, merged_config)

        client = self.client.parent if is_parent else self.client.child
        client.logger.info(
            'tx_config=%s, is_parent=%s, is_write=%s', tx_config, is_parent, is_write
        )

        def estimate_gas(config: ITransactionRequestConfig) -> int:
            if method:
                config.pop('value', None)  # already registered on method => ignored
                return method.estimate_gas(config)
            else:
                return client.estimate_gas(config)

        if not is_write:
            return tx_config

        is_eip_1559_supported = self.client.is_eip_1559_supported(is_parent)
        is_max_fee_provided = tx_config.get('max_fee_per_gas') or tx_config.get(
            'max_priority_fee_per_gas'
        )

        if not is_eip_1559_supported and is_max_fee_provided:
            raise EIP1559NotSupportedException

        if not tx_config.get('gas_limit'):
            tx_config['gas_limit'] = estimate_gas(tx_config)

        if not tx_config.get('nonce'):
            tx_config['nonce'] = client.get_transaction_count(
                tx_config['from'], 'pending'
            )
        tx_config.setdefault('chain_id', client.chain_id)

        return tx_config

    def transfer_erc_20(
        self,
        to: str,
        amount: int,
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Transfer ERC-20 token."""
        method = self.contract.method('transfer', to, amount)
        return self.process_write(method, option, private_key)

    def transfer_erc_721(
        self,
        from_: str,
        to: str,
        token_id: int,
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Transfer ERC-721 token."""
        method = self.contract.method('transferFrom', from_, to, token_id)
        return self.process_write(method, option, private_key)

    def check_for_root(self) -> None:
        """Assert that method is called on parent chain instance."""
        if not self.is_parent:
            raise AllowedOnRootException

    def check_for_child(self) -> None:
        """Assert that method is called on child chain instance."""
        if self.is_parent:
            raise AllowedOnChildException

    def transfer_erc_1155(
        self,
        from_: str,
        to: str,
        amount: int,
        token_id: int,
        data: bytes | None = b'',
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Transfer ERC-1155 token."""
        method = self.contract.method(
            'safeTransferFrom',
            from_,
            to,
            token_id,
            amount,
            data or b'',
        )
        return self.process_write(method, option, private_key)



from __future__ import annotations

from matic.json_types import ITransactionOption
from matic.utils.base_token import BaseToken
from matic.utils.web3_side_chain_client import Web3SideChainClient


class RootChainManager(BaseToken):
    """Root chain manager handles common operations related to POS bridge withdrawal."""

    def __init__(self, client: Web3SideChainClient, address: str) -> None:
        super().__init__(
            address=address,
            name='RootChainManager',
            bridge_type='pos',
            is_parent=True,
            client=client,
        )

    def deposit(
        self,
        user_address: str,
        token_address: str,
        deposit_data: bytes,
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Finish deposit operation.

        When token is approved, this is used to process "conversion". It may take 5-10
        minutes after this call to receive funds.

        Args:
            user_address: Receiving user address
            token_address: Address of contract defining token
            deposit_data: Pre-built binary data to send
            private_key: Receiving user PK
            option: Standard transaction option.
        """
        method = self.method('depositFor', user_address, token_address, deposit_data)
        return self.process_write(method, option, private_key)

    def exit(  # noqa: A003
        self,
        exit_payload: bytes,
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Finish exit operation.

        Args:
            exit_payload: Pre-built binary data to send
            private_key: Sender PK
            option: Standard transaction option.
        """
        method = self.method('exit', exit_payload)
        return self.process_write(method, option, private_key)

    def is_exit_processed(self, exit_hash: bytes) -> bool:
        """Check if exit was already processed for given transaction."""
        method = self.method('processedExits', exit_hash)
        return self.process_read(method)



from __future__ import annotations

from typing import Any, Iterable

import sha3  # pysha3

from matic.abstracts import BaseWeb3Client
from matic.web3_client import Web3Client as Web3ClientClass

__all__ = ['keccak256', 'resolve', 'UnstoppableDomains', 'Web3Client']


def keccak256(list_of_bytes: Iterable[bytes]) -> bytes:
    """Compute the sha3_256 flavor hash.

    Args:
        list_of_bytes: A list of bytes to be hashed.

    Returns:
        bytes: Hash value in :class:`bytes` (32 bytes).

    Raises:
        TypeError: If ``bytes`` or ``bytearray`` is used instead of sequence as input.
    """
    if isinstance(list_of_bytes, (bytes, bytearray)):  # type: ignore[unreachable]
        raise TypeError(
            f"Expected sequence of bytes or bytearray's, got: {type(list_of_bytes)}"
        )

    m = sha3.keccak_256()
    for item in list_of_bytes:
        m.update(item)

    return m.digest()


def resolve(obj: dict[str, Any], path: str | Iterable[str]):
    """Get value from nested dictionary by dotted path."""
    if isinstance(path, str):
        path = path.split('.')
    for key in path:
        obj = obj[key]
    return obj


UnstoppableDomains: dict[str, str] = {}

Web3Client: type[BaseWeb3Client] = Web3ClientClass
"""This can be assigned to use any other client class."""



from __future__ import annotations

from typing import Iterable, Sequence

from matic.constants import LogEventSignature
from matic.json_types import IExitTransactionOption, ITransactionOption
from matic.pos.pos_token import TokenWithApproveAll


class ERC1155(TokenWithApproveAll):
    """Arbitrary ERC-1155-compliant token."""

    CONTRACT_NAME: str = 'ChildERC1155'
    BURN_EVENT_SIGNATURE: bytes = LogEventSignature.ERC_1155_TRANSFER

    @property
    def mintable_predicate_address(self) -> str | None:
        """Address of mintable predicte for this token."""
        return getattr(
            self.client.config, 'erc_1155_mintable_predicate', ''
        ) or self.client.get_config('Main.POSContracts.MintableERC1155PredicateProxy')

    def get_balance(
        self, user_address: str, token_id: int, option: ITransactionOption | None = None
    ) -> int:
        """Get balance of a user for supplied token."""
        method = self.method('balanceOf', user_address, token_id)
        return self.process_read(method, option)

    def approve_all_for_mintable(
        self, private_key: str | None = None, option: ITransactionOption | None = None
    ):
        """Approve all tokens for mintable token."""
        self.check_for_root()
        address = self.mintable_predicate_address
        assert address
        return self._approve_all(address, private_key, option)

    def deposit(
        self,
        amount: int,
        token_id: int,
        user_address: str,
        data: bytes | None = None,
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Deposit supplied amount of token for a user."""
        self.check_for_root()
        return self.deposit_many(
            amounts=[amount],
            token_ids=[token_id],
            user_address=user_address,
            data=data or b'',
            private_key=private_key,
            option=option,
        )

    def deposit_many(
        self,
        amounts: Iterable[int],
        token_ids: Iterable[int],
        user_address: str,
        data: bytes | None = None,
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Deposit supplied amount of multiple token for user."""
        self.check_for_root()

        amount_in_abi = self.client.parent.encode_parameters(
            [
                token_ids,
                amounts,
                data or b'',
            ],
            ['uint256[]', 'uint256[]', 'bytes'],
        )

        return self.root_chain_manager.deposit(
            user_address,
            self.address,
            amount_in_abi,
            private_key,
            option,
        )

    def withdraw_start(
        self,
        token_id: int,
        amount: int,
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Start withdraw process by burning the required amount for a token."""
        self.check_for_child()

        method = self.method('withdrawSingle', token_id, amount)
        return self.process_write(method, option, private_key)

    def withdraw_start_many(
        self,
        token_ids: Sequence[int],
        amounts: Sequence[int],
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Start the withdraw process by burning multiple tokens at a time."""
        self.check_for_child()

        method = self.method('withdrawBatch', token_ids, amounts)
        return self.process_write(method, option, private_key)

    def withdraw_exit_many(
        self,
        burn_transaction_hash: bytes,
        private_key: str | None = None,
        option: IExitTransactionOption | None = None,
    ):
        """Exit the multiple withdraw process.

        Exit the withdraw process for many burned transaction and get the
        burned amount on root chain.

        This function fetches blocks and builds a proof manually.
        """
        return self.withdraw_exit_pos(
            burn_transaction_hash,
            LogEventSignature.ERC_1155_BATCH_TRANSFER,
            private_key,
            False,
            option=option,
        )

    def withdraw_exit_faster_many(
        self,
        burn_transaction_hash: bytes,
        private_key: str | None = None,
        option: IExitTransactionOption | None = None,
    ):
        """Exit the multiple withdraw process.

        Exit the withdraw process for many burned transaction and get the
        burned amount on root chain.

        This function uses API to get proof faster.
        """
        return self.withdraw_exit_pos(
            burn_transaction_hash,
            LogEventSignature.ERC_1155_BATCH_TRANSFER,
            private_key,
            True,
            option=option,
        )

    def is_withdraw_exited_many(self, tx_hash: bytes) -> bool:
        """Check if batch exit has been completed for a transaction hash."""
        return self.is_withdrawn(tx_hash, LogEventSignature.ERC_1155_BATCH_TRANSFER)

    def transfer(
        self,
        from_: str,
        to: str,
        amount: int,
        token_id: int,
        data: bytes | None = b'',
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Transfer the required amount of a token to another user."""
        return self.transfer_erc_1155(
            from_, to, amount, token_id, data, private_key, option
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Final

import rlp

from matic import services
from matic.abstracts import BaseWeb3Client
from matic.constants import LogEventSignature
from matic.exceptions import BurnTxNotCheckPointedException, ProofAPINotSetException
from matic.json_types import IBaseClientConfig, IRootBlockInfo, ITransactionReceipt
from matic.pos.root_chain import RootChain
from matic.utils import proof_utils
from matic.utils.web3_side_chain_client import Web3SideChainClient

ERC_721_HASHES: Final = {
    LogEventSignature.ERC_721_TRANSFER,
    LogEventSignature.ERC_721_TRANSFER_WITH_METADATA,
}
ERC_1155_HASHES: Final = {
    LogEventSignature.ERC_1155_TRANSFER,
    LogEventSignature.ERC_1155_BATCH_TRANSFER,
}
ZERO_SIG: Final = bytes(32)


@dataclass
class _IChainBlockInfo:
    """Internal dataclass."""

    last_child_block: int
    tx_block_number: int


class ExitUtil:
    """Helper utility class for building and performing exit actions with POS bridge."""

    _matic_client: BaseWeb3Client
    root_chain: RootChain
    config: IBaseClientConfig

    def __init__(self, client: Web3SideChainClient, root_chain: RootChain):
        self._matic_client = client.child
        self.root_chain = root_chain
        self.config = client.config

    def _get_log_index(self, log_event_sig: bytes, receipt: ITransactionReceipt) -> int:
        log_index = None

        if log_event_sig in ERC_721_HASHES:
            log_index = next(
                (
                    i
                    for i, log in enumerate(receipt.logs)
                    if (
                        len(log.topics) >= 2
                        and log.topics[0].lower() == log_event_sig.lower()
                        and log.topics[2].lower() == ZERO_SIG
                    )
                ),
                None,
            )
        elif log_event_sig in ERC_1155_HASHES:
            log_index = next(
                (
                    i
                    for i, log in enumerate(receipt.logs)
                    if (
                        len(log.topics) >= 3
                        and log.topics[0].lower() == log_event_sig.lower()
                        and log.topics[3].lower() == ZERO_SIG
                    )
                ),
                None,
            )
        else:
            log_index = next(
                (
                    i
                    for i, log in enumerate(receipt.logs)
                    if (log.topics and log.topics[0].lower() == log_event_sig.lower())
                ),
                None,
            )

        if log_index is None:
            raise ValueError('Log not found in receipt')
        return log_index

    def _get_all_log_indices(
        self, log_event_sig: bytes, receipt: ITransactionReceipt
    ) -> list[int]:
        if log_event_sig in ERC_721_HASHES:
            log_indices = [
                i
                for i, log in enumerate(receipt.logs)
                if (
                    len(log.topics) >= 2
                    and log.topics[0].lower() == log_event_sig.lower()
                    and log.topics[2].lower() == ZERO_SIG
                )
            ]
        elif log_event_sig in ERC_1155_HASHES:
            log_indices = [
                i
                for i, log in enumerate(receipt.logs)
                if (
                    len(log.topics) >= 3
                    and log.topics[0].lower() == log_event_sig.lower()
                    and log.topics[3].lower() == ZERO_SIG
                )
            ]
        elif log_event_sig == LogEventSignature.ERC_721_BATCH_TRANSFER:
            log_indices = [
                i
                for i, log in enumerate(receipt.logs)
                if (
                    len(log.topics) >= 2
                    and log.topics[0].lower() == LogEventSignature.ERC_20_TRANSFER
                    and log.topics[2].lower() == ZERO_SIG
                )
            ]
        else:
            log_indices = [
                i
                for i, log in enumerate(receipt.logs)
                if log.topics and log.topics[0].lower() == log_event_sig.lower()
            ]

        if not log_indices:
            raise ValueError('Log not found in receipt')

        return log_indices

    def get_chain_block_info(self, burn_tx_hash: bytes) -> _IChainBlockInfo:
        """Obtain information about block that includes given transaction."""
        tx = self._matic_client.get_transaction(burn_tx_hash)
        tx_block = tx.block_number
        assert tx_block is not None

        return _IChainBlockInfo(
            last_child_block=self.root_chain.last_child_block,
            tx_block_number=tx_block,
        )

    def _is_checkpointed(self, data: _IChainBlockInfo) -> bool:
        return int(data.last_child_block) >= int(data.tx_block_number)

    def is_checkpointed(self, burn_tx_hash: bytes) -> bool:
        """Check if given transaction is checkpointed."""
        return self._is_checkpointed(self.get_chain_block_info(burn_tx_hash))

    def _get_root_block_info(self, tx_block_number: int) -> IRootBlockInfo:
        """Returns info about block int existence on parent chain."""
        # find in which block child was included in parent
        block_number = self.root_chain.find_root_block_from_child(tx_block_number)
        _, start, end, _, _ = self.root_chain.method(
            'headerBlocks', block_number
        ).read()

        return IRootBlockInfo(
            header_block_number=block_number,
            start=start,
            end=end,
        )

    def _get_root_block_info_from_api(self, tx_block_number: int) -> IRootBlockInfo:
        try:
            header_block = services.get_block_included(
                self.config['network'], tx_block_number
            )
            self._matic_client.logger.debug('block info from API %s', header_block)

            if not (
                header_block
                and header_block.start
                and header_block.end
                and header_block.header_block_number
            ):
                raise ValueError('Network API Error')
            return header_block
        except Exception as e:  # noqa
            self._matic_client.logger.error('Block info from API error: %r', e)
            return self._get_root_block_info(tx_block_number)

    def _get_block_proof(
        self, tx_block_number: int, root_block_info: IRootBlockInfo
    ) -> bytes:
        return proof_utils.build_block_proof(
            self._matic_client,
            int(root_block_info.start),
            int(root_block_info.end),
            tx_block_number,
        )

    def _get_block_proof_from_api(
        self, tx_block_number: int, root_block_info: IRootBlockInfo
    ) -> bytes:
        try:
            block_proof = services.get_proof(
                self.config['network'],
                root_block_info.start,
                root_block_info.end,
                tx_block_number,
            )
            if not block_proof:
                raise RuntimeError('Network API Error')

            self._matic_client.logger.debug('block proof from API 1')
            return block_proof
        except ProofAPINotSetException:
            raise
        except Exception as e:  # noqa
            self._matic_client.logger.error('API error: %r', e)
            return self._get_block_proof(tx_block_number, root_block_info)

    def build_payload_for_exit(
        self, burn_tx_hash: bytes, index: int, log_event_sig: bytes, is_fast: bool
    ) -> bytes:
        """Build exit payload for transaction hash."""
        if index < 0:
            raise ValueError('Index must not be a negative integer')

        def get_indices(
            log_event_sig: bytes, receipt: ITransactionReceipt
        ) -> list[int]:
            if index > 0:
                log_indices = self._get_all_log_indices(log_event_sig, receipt)
                if index >= len(log_indices):
                    raise ValueError(
                        'Index is greater than the number of tokens in this transaction'
                    )
                return [log_indices[index]]
            else:
                return [self._get_log_index(log_event_sig, receipt)]

        return self._build_multiple_payloads_for_exit(
            burn_tx_hash, log_event_sig, is_fast, get_indices
        )[0]

    def build_multiple_payloads_for_exit(
        self, burn_tx_hash: bytes, log_event_sig: bytes, is_fast: bool
    ) -> list[bytes]:
        """Build exit payload for multiple indices."""
        return self._build_multiple_payloads_for_exit(
            burn_tx_hash, log_event_sig, is_fast, self._get_all_log_indices
        )

    def _build_multiple_payloads_for_exit(
        self,
        burn_tx_hash: bytes,
        log_event_sig: bytes,
        is_fast: bool,
        get_indices: Callable[[bytes, ITransactionReceipt], list[int]],
    ) -> list[bytes]:
        if is_fast and not services.DEFAULT_PROOF_API_URL:
            raise ProofAPINotSetException

        block_info = self.get_chain_block_info(burn_tx_hash)

        if not self._is_checkpointed(block_info):
            raise ValueError('Burn transaction has not been checkpointed as yet')

        # step 1 - Get Block int from transaction hash
        tx_block_number = block_info.tx_block_number

        # step 2-  get transaction receipt from txhash and
        # block information from block int
        receipt = self._matic_client.get_transaction_receipt(burn_tx_hash)
        block = self._matic_client.get_block_with_transaction(tx_block_number)

        # step  3 - get information about block saved in parent chain
        if is_fast:
            root_block_info = self._get_root_block_info_from_api(tx_block_number)
        else:
            root_block_info = self._get_root_block_info(tx_block_number)

        # step 4 - build block proof
        if is_fast:
            block_proof = self._get_block_proof_from_api(
                tx_block_number, root_block_info
            )
        else:
            block_proof = self._get_block_proof(tx_block_number, root_block_info)

        # step 5- create receipt proof
        receipt_proof = proof_utils.get_receipt_proof(
            receipt, block, self._matic_client
        )
        log_indices = get_indices(log_event_sig, receipt)

        # step 6 - encode payloads, convert into hex
        return [
            self._encode_payload(
                root_block_info.header_block_number,
                block_proof,
                tx_block_number,
                block.timestamp,
                block.transactions_root,
                block.receipts_root,
                proof_utils.get_receipt_bytes(receipt),  # rlp encoded
                receipt_proof['parent_nodes'],
                receipt_proof['path'],
                log_index,
            )
            for log_index in log_indices
        ]

    def _encode_payload(
        self,
        header_number,
        block_proof: bytes,
        block_number,
        timestamp,
        transactions_root,
        receipts_root,
        receipt,
        receipt_parent_nodes,
        path,
        log_index,
    ) -> bytes:
        return rlp.encode(
            [
                header_number,
                block_proof,
                block_number,
                timestamp,
                transactions_root,
                receipts_root,
                receipt,
                rlp.encode(receipt_parent_nodes),
                b'\x00' + path,
                log_index,
            ]
        )

    def get_exit_hash(
        self, burn_tx_hash: bytes, index: int, log_event_sig: bytes
    ) -> bytes:
        """Build exit hash for burn transaction."""
        last_child_block = self.root_chain.last_child_block
        receipt = self._matic_client.get_transaction_receipt(burn_tx_hash)
        block_result = self._matic_client.get_block_with_transaction(
            receipt.block_number
        )
        block = block_result
        if not self._is_checkpointed(
            _IChainBlockInfo(
                last_child_block=last_child_block, tx_block_number=receipt.block_number
            )
        ):
            raise BurnTxNotCheckPointedException()

        receipt_proof = proof_utils.get_receipt_proof(
            receipt, block, self._matic_client
        )

        log_index = None
        nibble = b''.join(
            b''.join(
                (
                    (byte // 0x10).to_bytes(1, 'big'),
                    (byte % 0x10).to_bytes(1, 'big'),
                )
            )
            for byte in receipt_proof['path']
        )

        if index > 0:
            log_indices = self._get_all_log_indices(log_event_sig, receipt)
            log_index = log_indices[index]

        log_index = self._get_log_index(log_event_sig, receipt)

        return self._matic_client.etherium_sha3(
            ['uint256', 'bytes', 'uint256'],
            [receipt.block_number, nibble, log_index],
        )



from __future__ import annotations

from logging import getLogger
from typing import Any, Final

import matic.utils
from matic.abstracts import BaseWeb3Client
from matic.json_types import IBaseClientConfig
from matic.utils.abi_manager import ABIManager


class Web3SideChainClient:
    """Web3 client class for a side chain."""

    parent: BaseWeb3Client
    child: BaseWeb3Client
    abi_manager: ABIManager

    logger: Final = getLogger(__package__)

    def __init__(self, config: IBaseClientConfig) -> None:
        # config.parent.default_config = config.parent.default_config or {}
        # config.child.default_config = config.child.default_config or {}
        self.config = config

        web3_client_cls = matic.utils.Web3Client
        if not web3_client_cls:
            raise ValueError('web3_client_cls is not set')

        self.resolution = matic.utils.UnstoppableDomains or None

        self.parent = web3_client_cls(
            (config.get('parent') or {}).get('provider'), self.logger  # type: ignore
        )
        self.child = web3_client_cls(
            (config.get('child') or {}).get('provider'), self.logger  # type: ignore
        )

        try:
            self.abi_manager = ABIManager(config['network'], config['version'])
        except Exception as e:
            raise ValueError(
                f'network {config["network"]} - {config["version"]} is not supported'
            ) from e

    def get_abi(self, name: str, type_: str | None = None) -> dict[str, Any]:
        """Get ABI dictionary for given name and type."""
        return self.abi_manager.get_abi(name, type_)

    def get_config(self, path: str) -> Any:
        """Fetch configuration option by a dotted path."""
        return self.abi_manager.get_config(path)

    @property
    def main_plasma_contracts(self):
        """Parameters of plasma contracts."""
        return self.get_config('Main.Contracts')

    @property
    def main_pos_contracts(self):
        """Parameters of POS contracts."""
        return self.get_config('Main.POSContracts')

    def is_eip_1559_supported(self, is_parent: bool) -> bool:
        """Check if EIP-1559 (improved fee specification) is available for chain."""
        if is_parent:
            return self.get_config('Main.SupportsEIP1559')
        else:
            return self.get_config('Matic.SupportsEIP1559')



from __future__ import annotations

from typing import Any, Final

from typing_extensions import TypedDict

from matic.services import get_abi, get_address
from matic.utils import resolve


class _CacheItem(TypedDict):
    address: Any
    abi: dict[str, dict[str, Any]]


CACHE: dict[tuple[str, str], _CacheItem] = {}
DEFAULT_BRIDGE_TYPE: Final = 'plasma'


class ABIManager:
    """Caching manager for fetched contract ABI dicts."""

    def __init__(self, network_name: str, version: str) -> None:
        self.network_name, self.version = network_name, version

        addr = get_address(network_name, version)
        CACHE[(network_name, version)] = {
            'address': addr,
            'abi': {},
        }

    def get_config(self, path: str) -> Any:
        """Get option from cache item by dotted path."""
        return resolve(
            CACHE[(self.network_name, self.version)]['address'],
            path,
        )

    def get_abi(
        self, contract_name: str, bridge_type: str | None = None
    ) -> dict[str, Any]:
        """Get ABI dict for contract and memoise it."""
        bridge_type = bridge_type or DEFAULT_BRIDGE_TYPE
        abi = (
            CACHE[(self.network_name, self.version)]['abi']
            .get(bridge_type, {})
            .get(contract_name)
        )
        if abi is not None:
            return abi

        abi = get_abi(self.network_name, self.version, bridge_type, contract_name)
        self.set_abi(contract_name, bridge_type, abi)
        return abi

    def set_abi(
        self, contract_name: str, bridge_type: str, abi: dict[str, Any]
    ) -> None:
        """Store ABI dict in cache."""
        abi_store = CACHE[(self.network_name, self.version)]['abi']
        abi_store.setdefault(bridge_type, {})[contract_name] = abi



"""Conversion utils: web3 structures to matic."""

from __future__ import annotations

from typing import Any

from web3 import Web3

from matic.json_types import (
    ILog,
    ITransactionData,
    ITransactionReceipt,
    ITransactionRequestConfig,
)


def matic_tx_request_config_to_web3(data: ITransactionRequestConfig | None = None):
    """Transaction request: matic to web3."""
    config: dict[str, Any] = dict(data or {})
    type_ = config.get('type')
    prepared = {
        'chainId': config.get('chain_id'),
        'data': config.get('data'),
        'from': config.get('from'),
        'gas': config.get('gas_limit'),
        'gasPrice': config.get('gas_price'),
        'nonce': config.get('nonce'),
        'to': config.get('to'),
        'value': config.get('value'),
        'maxFeePerGas': config.get('max_fee_per_gas'),
        'maxPriorityFeePerGas': config.get('max_priority_fee_per_gas'),
        'type': Web3.toHex(type_) if type_ else None,
        'hardfork': config.get('hardfork'),
    }
    return {k: v for k, v in prepared.items() if v is not None}


def web3_log_to_matic_log(log: Any) -> ILog:
    """Log: web3 to matic."""
    return ILog(
        address=log.get('address'),
        data=log.get('data'),
        topics=log.get('topics'),
        log_index=log.get('logIndex'),
        transaction_hash=log.get('transactionHash'),
        transaction_index=log.get('transactionIndex'),
        block_hash=log.get('blockHash'),
        block_number=log.get('blockNumber'),
        removed=log.get('removed'),
    )


def web3_receipt_to_matic_receipt(receipt: Any) -> ITransactionReceipt:
    """Transaction receipt: web3 to matic."""
    return ITransactionReceipt(
        block_hash=receipt.get('blockHash'),
        block_number=receipt.get('blockNumber'),
        contract_address=receipt.get('contractAddress'),
        cumulative_gas_used=receipt.get('cumulativeGasUsed'),
        from_=receipt.get('from'),
        gas_used=receipt.get('gasUsed'),
        status=receipt.get('status'),
        to=receipt.get('to'),
        transaction_hash=receipt.get('transactionHash'),
        transaction_index=receipt.get('transactionIndex'),
        events=receipt.get('events', []),
        logs=list(map(web3_log_to_matic_log, receipt.get('logs', []))),
        logs_bloom=receipt.get('logsBloom'),
        root=receipt.get('root'),
        type=receipt.get('type'),
    )


def web3_tx_to_matic_tx(tx) -> ITransactionData:
    """Transaction: web3 to matic."""
    return ITransactionData(
        transaction_hash=tx.hash,
        nonce=tx.nonce,
        block_hash=tx.blockHash,
        block_number=tx.blockNumber,
        transaction_index=tx.transactionIndex,
        from_=tx['from'],
        to=tx.to,
        value=tx.value,
        gas_price=tx.gasPrice,
        gas=tx.gas,
        input=tx.input,
    )



from __future__ import annotations

from math import ceil, log2
from typing import Iterable, Sequence

from matic.utils import keccak256


class MerkleTree:
    """Hash tree.

    See https://en.wikipedia.org/wiki/Merkle_tree
    """

    leaves: list[bytes]
    layers: list[list[bytes]]

    def __init__(self, leaves: Sequence[bytes] | None = None):
        if not leaves:
            raise ValueError('At least 1 leaf needed')

        depth = self.estimate_depth(len(leaves))
        if depth > 20:
            raise ValueError('Depth must be 20 or less')

        self.leaves = [
            *leaves,
            *[bytes(32) for _ in range(2**depth - len(leaves))],
        ]

        self.layers = [self.leaves]
        self.create_hashes(self.leaves)

    @staticmethod
    def estimate_depth(size: int) -> int:
        """Estimate depth of a tree with given size.

        Like a heap, hash tree has height equal to ``log2(size)``.
        """
        return ceil(log2(size))

    @property
    def depth(self) -> int:
        """Tree depth."""
        return self.estimate_depth(len(self.leaves))

    def create_hashes(self, nodes: Sequence[bytes]) -> None:
        """Add nodes to tree."""
        if len(nodes) == 1:
            return

        tree_level = [
            keccak256([left, right]) for left, right in zip(nodes[::2], nodes[1::2])
        ]

        # is odd number of nodes
        if len(nodes) % 2:
            tree_level.append(nodes[-1])

        self.layers.append(tree_level)
        self.create_hashes(tree_level)

    @property
    def root(self) -> bytes:
        """Tree root."""
        return self.layers[-1][0]

    def get_proof(self, leaf: bytes) -> list[bytes]:
        """Get proof for leaf as a sequence of nodes."""
        index = next(
            (i for i, item in enumerate(self.leaves) if item == leaf),
            None,
        )

        if index is None:
            return []

        proof = []
        for layer in self.layers[:-1]:
            sibling_index = index + (-1 if index % 2 else 1)
            index //= 2
            proof.append(layer[sibling_index])

        return proof

    def verify(
        self, value: bytes, index: int, root: bytes, proof: Iterable[bytes]
    ) -> bool:
        """Verify given proof to match the tree."""
        if not isinstance(proof, Iterable) or not value or not root:
            return False

        hash_ = value
        for node in proof:
            if index % 2:
                hash_ = keccak256([node, hash_])
            else:
                hash_ = keccak256([hash_, node])

            index //= 2

        return hash_ == root



from __future__ import annotations

from typing import Final, final

MAX_AMOUNT: Final = 2**256 - 1
"""Max transaction amount."""


@final
class LogEventSignature:
    """Signatures of different transfer kinds."""

    ERC_20_TRANSFER: Final = bytes.fromhex(
        'ddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef'
    )
    ERC_721_TRANSFER: Final = bytes.fromhex(
        'ddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef'
    )
    ERC_1155_TRANSFER: Final = bytes.fromhex(
        'c3d58168c5ae7397731d063d5bbf3d657854427343f4c083240f7aacaa2d0f62'
    )
    ERC_721_BATCH_TRANSFER: Final = bytes.fromhex(
        'f871896b17e9cb7a64941c62c188a4f5c621b86800e3d15452ece01ce56073df'
    )
    ERC_1155_BATCH_TRANSFER: Final = bytes.fromhex(
        '4a39dc06d4c0dbc64b70af90fd698a233a518aa5d07e595d983b8c0526c8f7fb'
    )
    ERC_721_TRANSFER_WITH_METADATA: Final = bytes.fromhex(
        'f94915c6d1fd521cee85359239227480c7e8776d7caf1fc3bacad5c269b66a14'
    )
    STATE_SYNCED_EVENT: Final = bytes.fromhex(
        '103fed9db65eac19c4d870f49ab7520fe03b99f1838e5996caf47e9e43308392'
    )



from __future__ import annotations


class MaticException(Exception):
    """Base exception class for this library."""


class AllowedOnRootException(MaticException):
    """The action is allowed only on root token."""


class AllowedOnChildException(MaticException):
    """The action is allowed only on child token."""


class ProofAPINotSetException(MaticException):
    """Proof api is not set."""


class BurnTxNotCheckPointedException(MaticException):
    """Burn transaction has not been checkpointed as yet."""


class EIP1559NotSupportedException(MaticException):
    """The chain doesn't support EIP-1559."""


class NullSpenderAddressException(MaticException):
    """Please provide spender address."""



from __future__ import annotations

from typing import Callable

from matic.json_types import IExitTransactionOption, IPOSContracts, ITransactionOption
from matic.pos.exit_util import ExitUtil
from matic.pos.root_chain_manager import RootChainManager
from matic.utils.base_token import BaseToken
from matic.utils.web3_side_chain_client import Web3SideChainClient


class POSToken(BaseToken):
    """Base class for all tokens based on POS bridge protocol."""

    _predicate_address: str | None = None
    CONTRACT_NAME: str  # TODO: should be abstract
    BURN_EVENT_SIGNATURE: bytes  # TODO: should be abstract

    def __init__(
        self,
        token_address: str,
        is_parent: bool,
        client: Web3SideChainClient,
        get_pos_contracts: Callable[[], IPOSContracts],
    ) -> None:
        super().__init__(
            is_parent=is_parent,
            address=token_address,
            name=self.CONTRACT_NAME,
            bridge_type='pos',
            client=client,
        )
        self.get_pos_contracts = get_pos_contracts

    @property
    def root_chain_manager(self) -> RootChainManager:
        """Get RootChainManager instance."""
        return self.get_pos_contracts().root_chain_manager

    @property
    def exit_util(self) -> ExitUtil:
        """Get ExitUtil instance."""
        return self.get_pos_contracts().exit_util

    @property
    def predicate_address(self) -> str:
        """Memoise and get predicate address from root chain manager."""
        if self._predicate_address:
            return self._predicate_address

        token_type = self.root_chain_manager.method('tokenToType', self.address).read()

        if not token_type:
            raise ValueError('Invalid Token Type')

        predicate_address = self.root_chain_manager.method(
            'typeToPredicate', token_type
        ).read()

        self._predicate_address = predicate_address
        return predicate_address

    def is_withdrawn(self, tx_hash: bytes, event_signature: bytes) -> bool:
        """Check if transaction withdrawal was completed."""
        exit_hash = self.exit_util.get_exit_hash(tx_hash, 0, event_signature)
        return self.root_chain_manager.is_exit_processed(exit_hash)

    def is_withdrawn_on_index(
        self, tx_hash: bytes, index: int, event_signature: bytes
    ) -> bool:
        """Check if transaction withdrawal was completed on index."""
        exit_hash = self.exit_util.get_exit_hash(tx_hash, index, event_signature)
        return self.root_chain_manager.is_exit_processed(exit_hash)

    def withdraw_exit_pos(
        self,
        burn_tx_hash: bytes,
        event_signature: bytes,
        private_key: str | None = None,
        is_fast: bool = True,
        index: int = 0,
        *,
        option: IExitTransactionOption | None = None,
    ) -> None:
        """Base POS exit method, called by more specific implementations."""
        self.check_for_root()

        event_signature = (
            option.get('burn_event_signature') if option is not None else None
        ) or event_signature

        payload = self.exit_util.build_payload_for_exit(
            burn_tx_hash, index, event_signature, is_fast
        )
        return self.root_chain_manager.exit(payload, private_key, option)

    def withdraw_exit(
        self,
        burn_transaction_hash: bytes,
        private_key: str | None = None,
        option: IExitTransactionOption | None = None,
    ):
        """Complete withdraw process.

        This function should be called after checkpoint has been submitted
        for the block containing burn tx.

        This function fetches required blocks and builds proof using them.
        """
        return self.withdraw_exit_pos(
            burn_transaction_hash,
            self.BURN_EVENT_SIGNATURE,
            private_key,
            False,
            option=option,
        )

    def withdraw_exit_faster(
        self,
        burn_transaction_hash: bytes,
        private_key: str | None = None,
        option: IExitTransactionOption | None = None,
    ):
        """Complete withdraw process.

        This function should be called after checkpoint has been submitted
        for the block containing burn tx.

        This function uses API to fetch proof data.
        """
        return self.withdraw_exit_pos(
            burn_transaction_hash,
            self.BURN_EVENT_SIGNATURE,
            private_key,
            True,
            option=option,
        )

    def is_withdraw_exited(self, tx_hash: bytes) -> bool:
        """Check if exit has been completed for a transaction hash."""
        return self.is_withdrawn(tx_hash, self.BURN_EVENT_SIGNATURE)


class TokenWithApproveAll(POSToken):
    """This is a general token with common methods for ERC721 and ERC1155."""

    def is_approved_all(
        self, user_address: str, option: ITransactionOption | None = None
    ) -> bool:
        """Check if a user is approved for all tokens."""
        self.check_for_root()
        method = self.contract.method(
            'isApprovedForAll', user_address, self.predicate_address
        )
        return bool(self.process_read(method, option))

    def _approve_all(
        self,
        predicate_address: str,
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        self.check_for_root()
        method = self.contract.method('setApprovalForAll', predicate_address, True)
        return self.process_write(method, option, private_key)

    def approve_all(
        self, private_key: str | None = None, option: ITransactionOption | None = None
    ):
        """Approve all tokens."""
        return self._approve_all(self.predicate_address, private_key, option)



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime
from logging import Logger
from typing import Any, Iterable, Sequence

from matic.json_types import (
    IBlock,
    IBlockWithTransaction,
    IJsonRpcRequestPayload,
    IJsonRpcResponse,
    ITransactionData,
    ITransactionReceipt,
    ITransactionRequestConfig,
    ITransactionWriteResult,
)


class BaseWeb3Client(ABC):
    def __init__(self, provider: Any, logger: Logger):
        self.provider = provider
        self.logger = logger

    @abstractmethod
    def get_contract(self, address: str, abi: Any) -> BaseContract:
        ...

    @abstractmethod
    def read(self, config: ITransactionRequestConfig) -> str:
        ...

    @abstractmethod
    def write(self, config: ITransactionRequestConfig) -> ITransactionWriteResult:
        ...

    @property
    @abstractmethod
    def gas_price(self) -> int:
        ...

    @abstractmethod
    def estimate_gas(self, config: ITransactionRequestConfig) -> int:
        ...

    @property
    @abstractmethod
    def chain_id(self) -> int:
        ...

    @abstractmethod
    def get_transaction_count(self, address: str, block_number: Any) -> int:
        ...

    @abstractmethod
    def get_transaction(self, transaction_hash: bytes) -> ITransactionData:
        ...

    @abstractmethod
    def get_transaction_receipt(self, transaction_hash: bytes) -> ITransactionReceipt:
        ...

    # @abstractmethod
    # def extend(property: str, methods: IMethod[])

    @abstractmethod
    def get_block(self, block_hash_or_block_number) -> IBlock:
        ...

    @abstractmethod
    def get_block_with_transaction(
        self, block_hash_or_block_number
    ) -> IBlockWithTransaction:
        ...

    def get_root_hash(self, start_block: int, end_block: int) -> bytes:
        return bytes.fromhex(
            self.send_rpc_request(
                {
                    'method': 'eth_getRootHash',
                    'params': [int(start_block), int(end_block)],
                    'id': round(datetime.now().timestamp()),
                }
            )['result']
        )

    @abstractmethod
    def send_rpc_request(self, request: IJsonRpcRequestPayload) -> IJsonRpcResponse:
        ...

    @abstractmethod
    def encode_parameters(self, params: Sequence[Any], types: Sequence[Any]) -> bytes:
        ...

    @abstractmethod
    def decode_parameters(self, encoded: bytes, types: Sequence[Any]) -> Sequence[Any]:
        ...

    @abstractmethod
    def etherium_sha3(self, types: Iterable[str], values: Iterable[Any]) -> bytes:
        ...


class BaseContractMethod(ABC):
    def __init__(self, address: str, logger: Logger, method: Any) -> None:
        self.address = address
        self.logger = logger
        self.method = method

    @abstractmethod
    def read(self, tx: ITransactionRequestConfig | None = None) -> Any:
        ...

    @abstractmethod
    def write(
        self, tx: ITransactionRequestConfig, private_key: str | None = None
    ) -> ITransactionWriteResult:
        ...

    @abstractmethod
    def estimate_gas(self, tx: ITransactionRequestConfig) -> int:
        ...

    @abstractmethod
    def encode_abi(self) -> Any:
        ...


class BaseContract(ABC):
    def __init__(self, address: str, logger: Logger):
        self.address = address
        self.logger = logger

    @abstractmethod
    def method(self, method_name: str, *args: Any) -> BaseContractMethod:
        ...



from __future__ import annotations

from matic.constants import LogEventSignature
from matic.json_types import IBaseClientConfig
from matic.pos.exit_util import ExitUtil
from matic.utils.base_token import BaseToken
from matic.utils.web3_side_chain_client import Web3SideChainClient


class BridgeClient:
    """Base class for POS bridge with reusable methods."""

    client: Web3SideChainClient
    exit_util: ExitUtil

    def __init__(self, config: IBaseClientConfig):
        self.client = Web3SideChainClient(config)

    def is_checkpointed(self, tx_hash: bytes) -> bool:
        """Check if transaction is checkpointed."""
        return self.exit_util.is_checkpointed(tx_hash)

    def is_deposited(self, deposit_tx_hash: bytes) -> bool:
        """Check if deposit has finished after exit (StateSynced happened)."""
        client = self.client

        token = BaseToken(
            address=client.abi_manager.get_config(
                'Matic.GenesisContracts.StateReceiver'
            ),
            is_parent=False,
            name='StateReceiver',
            bridge_type='genesis',
            client=client,
        )

        receipt = client.parent.get_transaction_receipt(deposit_tx_hash)
        last_state_id = token.process_read(token.contract.method('lastStateId'))

        event_signature = LogEventSignature.STATE_SYNCED_EVENT
        try:
            target_log = next(
                log for log in receipt.logs if log.topics[0] == event_signature
            )
        except StopIteration:
            raise RuntimeError('StateSynced event not found')

        root_state_id = client.child.decode_parameters(
            target_log.topics[1], ['uint256']
        )[0]
        return last_state_id >= root_state_id



from __future__ import annotations

from typing import Iterable

import rlp
from mpt import MerklePatriciaTrie as Trie

from matic.abstracts import BaseWeb3Client
from matic.json_types import IBaseBlock, IBlockWithTransaction, ITransactionReceipt
from matic.utils import keccak256
from matic.utils.merkle_tree import MerkleTree
from matic.utils.polyfill import removeprefix

# Implementation adapted from Tom French's `matic-proofs` library used under MIT License
# https://github.com/TomAFrench/matic-proofs


def get_fast_merkle_proof(
    web3: BaseWeb3Client,
    block_number: int,
    start_block: int,
    end_block: int,
) -> list[bytes]:
    """Get fast Merkle proof for block."""
    tree_depth = MerkleTree.estimate_depth(end_block - start_block + 1)

    # We generate the proof root down, whereas we need from leaf up
    reversed_proof: list[bytes] = []

    offset = start_block
    target_index = block_number - offset
    left_bound = 0
    right_bound = end_block - offset

    for depth in range(tree_depth):
        n_leaves = 2 ** (tree_depth - depth)

        # The pivot leaf is the last leaf which is included in the left subtree
        pivot_leaf = left_bound + n_leaves // 2 - 1

        if target_index > pivot_leaf:
            # Get the root hash to the merkle subtree to the left
            new_left_bound = pivot_leaf + 1
            # eslint-disable-next-line no-await-in-loop
            subtree_merkle_root = query_root_hash(
                web3, offset + left_bound, offset + pivot_leaf
            )
            reversed_proof.append(subtree_merkle_root)
            left_bound = new_left_bound
        else:
            # Things are more complex when querying to the right.
            # Root hash may come some layers down so we need to build a full tree
            # by padding with zeros.
            # Some trees may be completely empty.
            new_right_bound = min(right_bound, pivot_leaf)

            # Expect the tree to have a height one less than the current layer
            expected_height = tree_depth - (depth + 1)

            if right_bound <= pivot_leaf:
                # Tree is empty so we repeatedly hash zero to correct height
                subtree_merkle_root = recursive_zero_hash(expected_height, web3)
                reversed_proof.append(subtree_merkle_root)
            else:
                # Height of tree given by RPC node
                sub_tree_height = MerkleTree.estimate_depth(right_bound - pivot_leaf)

                # Find the difference in height between this and the subtree we want
                height_difference = expected_height - sub_tree_height

                # For every extra layer we need to fill 2*n leaves filled with the
                # merkle root of a zero-filled Merkle tree
                # We need to build a tree which has height_difference layers

                # The first leaf will hold the root hash as returned by the RPC
                # eslint-disable-next-line no-await-in-loop
                remaining_nodes_hash = query_root_hash(
                    web3, offset + pivot_leaf + 1, offset + right_bound
                )

                # The remaining leaves will hold the merkle root of a zero-filled
                # tree of height sub_tree_height
                leaf_roots = recursive_zero_hash(sub_tree_height, web3)

                # Build a merkle tree of correct size for the subtree using
                # these merkle roots
                leaves = [bytes(leaf_roots) for _ in range(2**height_difference)]
                leaves[0] = remaining_nodes_hash
                subtree_merkle_root = MerkleTree(leaves).root
                reversed_proof.append(subtree_merkle_root)

            right_bound = new_right_bound

    return reversed_proof[::-1]


def build_block_proof(
    matic_web3: BaseWeb3Client,
    start_block: int,
    end_block: int,
    block_number: int,
) -> bytes:
    """Get proof for block as single byte string."""
    proof = get_fast_merkle_proof(matic_web3, block_number, start_block, end_block)
    return b''.join(proof)


def query_root_hash(client: BaseWeb3Client, start_block: int, end_block: int) -> bytes:
    """Get root hash."""
    return client.get_root_hash(start_block, end_block)


def recursive_zero_hash(n: int, client: BaseWeb3Client) -> bytes:
    """Get n-th recursive zero hash."""
    if n == 0:
        return bytes(32)

    sub_hash = recursive_zero_hash(n - 1, client)
    return keccak256([client.encode_parameters([sub_hash] * 2, ['bytes32'] * 2)])


def get_receipt_proof(
    receipt: ITransactionReceipt,
    block: IBlockWithTransaction,
    web3: BaseWeb3Client,
    request_concurrency: int | None = None,
    receipts_val: Iterable[ITransactionReceipt] | None = None,
):
    """Get proof for receipt."""
    state_sync_tx_hash = get_state_sync_tx_hash(block).hex()
    receipts_trie = Trie({})

    receipts = receipts_val or [  # TODO: can be done in parallel
        web3.get_transaction_receipt(tx.transaction_hash)
        for tx in block.transactions
        if tx.transaction_hash != state_sync_tx_hash
    ]

    for sibling in receipts:
        path = rlp.encode(sibling.transaction_index)
        raw_receipt = get_receipt_bytes(sibling)
        receipts_trie.update(path, raw_receipt)

    stack = list(receipts_trie.find_path(rlp.encode(receipt.transaction_index)))
    node = stack[-1]

    return {
        'block_hash': receipt.block_hash,
        'parent_nodes': [n.raw() for n in stack],
        'root': block.receipts_root,
        'path': rlp.encode(receipt.transaction_index),
        'value': (node.data if is_typed_receipt(receipt) else rlp.decode(node.data)),
    }


def is_typed_receipt(receipt: ITransactionReceipt) -> bool:
    """Check if transaction was performed and type is non-zero."""
    return bool(receipt.status is not None and receipt.type not in {'0x0', '0x'})


def get_state_sync_tx_hash(block: IBaseBlock) -> bytes:
    """Get block's tx hash for state-sync receipt.

    Bor blockchain includes extra receipt/tx for state-sync logs,
    but it is not included in transactionRoot or receiptRoot.
    So, while calculating proof, we have to exclude them.

    This is derived from block's hash and int.
    """
    return keccak256(
        [
            # prefix for bor receipt
            b'matic-bor-receipt-',
            block.number.to_bytes(8, 'big'),  # 8 bytes of block number (BigEndian)
            block.hash,  # block hash
        ]
    )


def get_receipt_bytes(receipt: ITransactionReceipt) -> bytes:
    """Get binary representation of receipt for storing in trie."""
    encoded_data = rlp.encode(
        [
            (
                (b'\x01' if receipt.status else b'')
                if receipt.status is not None
                else receipt.root
            ),
            receipt.cumulative_gas_used,
            receipt.logs_bloom,
            # encoded log array
            [
                # [address, [topics array], data]
                [
                    bytes.fromhex(removeprefix(log.address, '0x')),
                    log.topics,
                    bytes.fromhex(removeprefix(log.data, '0x')),
                ]
                for log in receipt.logs
            ],
        ]
    )
    if is_typed_receipt(receipt):
        encoded_data = int(receipt.type, 0).to_bytes(1, 'big') + encoded_data

    return encoded_data


# def get_raw_header( block):
#     raise NotImplementedError



from __future__ import annotations

from logging import Logger
from typing import Any, Iterable, Literal, Sequence, cast

from eth_abi import decode_abi, encode_abi
from hexbytes.main import HexBytes
from web3 import Web3
from web3.contract import Contract
from web3.method import Method
from web3.providers.base import BaseProvider
from web3.types import RPCEndpoint

from matic.abstracts import BaseContract, BaseContractMethod, BaseWeb3Client
from matic.json_types import (
    IBlock,
    IBlockWithTransaction,
    IJsonRpcRequestPayload,
    ITransactionData,
    ITransactionReceipt,
    ITransactionRequestConfig,
    ITransactionWriteResult,
)
from matic.utils.polyfill import removeprefix
from matic.web3_client.utils import (
    matic_tx_request_config_to_web3,
    web3_receipt_to_matic_receipt,
    web3_tx_to_matic_tx,
)

__all__ = ['TransactionWriteResult', 'EthMethod', 'Web3Contract', 'Web3Client']


class TransactionWriteResult(ITransactionWriteResult):
    """Result of any writing call."""

    def __init__(self, tx_hash: bytes, client):
        self.tx_hash = tx_hash
        self.client = client

    def get_receipt(self, timeout: int = 120):
        """Get transaction receipt.

        Args:
            timeout: max seconds to wait.
        """
        receipt = self.client._web3.eth.wait_for_transaction_receipt(
            self.tx_hash, timeout=timeout
        )
        return web3_receipt_to_matic_receipt(receipt)

    @property
    def transaction_hash(self) -> bytes:
        """Hash of performed transaction."""
        return self.tx_hash


class EthMethod(BaseContractMethod):
    """Wrapper around web3 contract method (:class:`web3.method.Method`)."""

    def __init__(
        self, address: str, logger: Logger, method: Method[Any], client: Web3Client
    ) -> None:
        super().__init__(address, logger, method)
        self.method = method
        self.address = address
        self.client = client

    def read(self, tx: ITransactionRequestConfig | None = None):
        """Perform a read operation.

        This does not sign a transaction and does not affect the chain.
        """
        self.logger.debug('sending tx with config %s', tx)
        return self.method.call(matic_tx_request_config_to_web3(tx))

    def write(
        self,
        tx: ITransactionRequestConfig,
        private_key: str | None = None,
    ) -> TransactionWriteResult:
        """Perform a write operation.

        Transaction is signed (with given PK), affects the chain.
        """
        web3_tx = matic_tx_request_config_to_web3(tx)

        if private_key:
            tx_prep = self.method.build_transaction(web3_tx)
            self.logger.debug('Prepared tx: ', tx_prep)
            tx_signed = self.client._web3.eth.account.sign_transaction(
                tx_prep, private_key
            )
            res = self.client._web3.eth.send_raw_transaction(tx_signed.rawTransaction)
        else:
            res = self.client._web3.eth.send_transaction(web3_tx)

        return TransactionWriteResult(res, self.client)

    def estimate_gas(self, tx: ITransactionRequestConfig) -> int:
        """Estimate gas for given transaction.

        Warning:
            This method may fail if your transaction is invalid or cannot be executed.
        """
        return self.method.estimate_gas(matic_tx_request_config_to_web3(tx))

    def encode_abi(self) -> bytes:
        """Encode args according to method ABI and prepend the selector."""
        abi = [e['type'] for e in self.method.abi['inputs']]
        return bytes.fromhex(removeprefix(self.method.selector, '0x')) + encode_abi(
            abi, self.method.args
        )


class Web3Contract(BaseContract):
    """A wrapper around web3 contract (:class:`web3.contract.Contract`)."""

    def __init__(
        self, address: str, contract: Contract, logger: Logger, client: Web3Client
    ):
        super().__init__(address, logger)
        self.contract = contract
        self.client = client

    def method(self, method_name: str, *args: Any) -> EthMethod:
        """Obtain a method object by name and call arguments."""
        self.logger.debug('method_name %s; args method %s', method_name, args)
        return EthMethod(
            self.address,
            self.logger,
            self.contract.get_function_by_name(method_name)(*args),
            self.client,
        )


class Web3Client(BaseWeb3Client):
    """Implementation of web3 client."""

    _web3: Web3

    def __init__(self, provider: BaseProvider, logger: Logger):
        from web3.middleware import geth_poa_middleware

        super().__init__(provider, logger)
        self._web3 = Web3(provider)
        self._web3.middleware_onion.inject(geth_poa_middleware, layer=0)

    def read(self, config: ITransactionRequestConfig):
        """Perform a reading (non-modifying) operation."""
        return self._web3.eth.call(matic_tx_request_config_to_web3(config))

    def write(self, config: ITransactionRequestConfig):
        """Perform a writing (modifying) operation."""
        return TransactionWriteResult(
            self._web3.eth.send_transaction(matic_tx_request_config_to_web3(config)),
            self,
        )

    def get_contract(self, address: str, abi: dict[str, Any]) -> Web3Contract:
        """Obtain a contract from deployment address and ABI dictionary."""
        cont = self._web3.eth.contract(abi=abi, address=cast(Any, address))
        return Web3Contract(address, cont, self.logger, self)

    @property
    def gas_price(self) -> int:
        """Current gas price."""
        return self._web3.eth.gas_price

    def estimate_gas(self, transaction: ITransactionRequestConfig) -> int:
        """Estimate gas amount for transaction."""
        return self._web3.eth.estimate_gas(matic_tx_request_config_to_web3(transaction))

    def get_transaction_count(self, address: str, block_number: int) -> int:
        """Get amount of transactions in specified block for given address."""
        return self._web3.eth.get_transaction_count(address, block_number)

    @property
    def chain_id(self):
        """Current chain id."""
        return self._web3.eth.chain_id

    def _ensure_transaction_not_null(self, data) -> None:
        if not data:
            raise ValueError(
                'Could not retrieve transaction.'
                ' Either it is invalid or might be in archive node.'
            )

    def get_transaction(self, transaction_hash: bytes) -> ITransactionData:
        """Obtain transaction object by hash."""
        data = self._web3.eth.get_transaction(HexBytes(transaction_hash))
        self._ensure_transaction_not_null(data)
        return web3_tx_to_matic_tx(data)

    def get_transaction_receipt(
        self, transaction_hash: bytes, timeout: int = 120
    ) -> ITransactionReceipt:
        """Get receipt for transaction."""
        data = self._web3.eth.wait_for_transaction_receipt(
            HexBytes(transaction_hash), timeout
        )
        self._ensure_transaction_not_null(data)
        return web3_receipt_to_matic_receipt(data)

    def get_block(
        self,
        block_hash_or_block_number: Literal['latest', 'earliest', 'pending']
        | bytes
        | HexBytes
        | int,
    ) -> IBlock:
        """Get block (with raw transaction data) by hash or number."""
        if isinstance(block_hash_or_block_number, bytes):
            block_hash_or_block_number = HexBytes(block_hash_or_block_number)
        data: Any = self._web3.eth.get_block(block_hash_or_block_number)
        return IBlock(
            size=data.size,
            difficulty=data.difficulty,
            total_difficulty=data.totalDifficulty,
            uncles=data.uncles,
            number=data.number,
            hash=data.hash,
            parent_hash=data.parentHash,
            nonce=data.nonce,
            sha3_uncles=data.sha3Uncles,
            logs_bloom=data.logsBloom,
            transactions_root=data.transactionsRoot,
            state_root=data.stateRoot,
            receipts_root=data.receiptsRoot,
            miner=data.miner,
            extra_data=data.proofOfAuthorityData,
            gas_limit=data.gasLimit,
            gas_used=data.gasUsed,
            timestamp=data.timestamp,
            # base_fee_per_gas=data.baseFeePerGas,
            transactions=data.transactions,
        )

    def get_block_with_transaction(
        self,
        block_hash_or_block_number: Literal['latest', 'earliest', 'pending']
        | bytes
        | HexBytes
        | int,
    ) -> IBlockWithTransaction:
        """Get block (with decoded transaction data) by hash or number."""
        if isinstance(block_hash_or_block_number, bytes):
            block_hash_or_block_number = HexBytes(block_hash_or_block_number)
        data: Any = self._web3.eth.get_block(block_hash_or_block_number, True)
        return IBlockWithTransaction(
            size=data.size,
            difficulty=data.difficulty,
            total_difficulty=data.totalDifficulty,
            uncles=data.uncles,
            number=data.number,
            hash=data.hash,
            parent_hash=data.parentHash,
            nonce=data.nonce,
            sha3_uncles=data.sha3Uncles,
            logs_bloom=data.logsBloom,
            transactions_root=data.transactionsRoot,
            state_root=data.stateRoot,
            receipts_root=data.receiptsRoot,
            miner=data.miner,
            extra_data=data.proofOfAuthorityData,
            gas_limit=data.gasLimit,
            gas_used=data.gasUsed,
            timestamp=data.timestamp,
            # base_fee_per_gas=data.baseFeePerGas,
            transactions=list(map(web3_tx_to_matic_tx, data.transactions)),
        )

    def send_rpc_request(self, request: IJsonRpcRequestPayload):
        """Perform arbitrary RPC request."""
        return self._web3.provider.make_request(
            cast(RPCEndpoint, request['method']), request['params']
        )

    def encode_parameters(self, params: Sequence[Any], types: Sequence[Any]):
        """Encode ABI parameters according to schema."""
        return encode_abi(types, params)

    def decode_parameters(self, binary_data: bytes, types: Sequence[Any]):
        """Decode binary data to ABI parameters according to schema."""
        return decode_abi(types, binary_data)

    def etherium_sha3(self, types: Iterable[str], values: Iterable[Any]) -> bytes:
        """Calculate solidity keccak hash of given values (encoded as types)."""
        return self._web3.solidityKeccak(types, values)



from __future__ import annotations

from matic.constants import MAX_AMOUNT, LogEventSignature
from matic.exceptions import NullSpenderAddressException
from matic.json_types import IAllowanceTransactionOption, ITransactionOption
from matic.pos.pos_token import POSToken


class ERC20(POSToken):
    """Arbitrary ERC-20-compliant token."""

    CONTRACT_NAME: str = 'ChildERC20'
    BURN_EVENT_SIGNATURE: bytes = LogEventSignature.ERC_20_TRANSFER

    def get_balance(
        self, user_address: str, option: ITransactionOption | None = None
    ) -> int:
        """Get balance of a user for supplied token."""
        method = self.method(
            'balanceOf',
            user_address,
        )
        return self.process_read(method, option)

    def get_allowance(
        self, user_address: str, option: IAllowanceTransactionOption | None = None
    ):
        """Get allowance of user."""
        predicate_address = (
            option['spender_address'] if option else None
        ) or self.predicate_address

        method = self.method(
            'allowance',
            user_address,
            predicate_address,
        )
        return self.process_read(method, option)

    def approve(
        self,
        amount: int,
        private_key: str | None = None,
        option: IAllowanceTransactionOption | None = None,
    ):
        """Approve specified amount to contract."""
        spender_address = option.get('spender_address') if option else None
        if not spender_address and not self.is_parent:
            raise NullSpenderAddressException

        predicate_address = (
            spender_address if spender_address else self.predicate_address
        )

        method = self.method('approve', predicate_address, amount)
        return self.process_write(method, option, private_key)

    def approve_max(
        self, private_key: str, option: IAllowanceTransactionOption | None = None
    ):
        """Approve max possible amount of token to contract."""
        return self.approve(MAX_AMOUNT, private_key, option)

    def deposit(
        self,
        amount: int,
        user_address: str,
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Deposit given amount of token for user."""
        self.check_for_root()
        amount_in_abi = self.client.parent.encode_parameters([amount], ['uint256'])
        return self.root_chain_manager.deposit(
            user_address,
            self.address,
            amount_in_abi,
            private_key,
            option,
        )

    def _deposit_ether(
        self,
        amount: int,
        user_address: str,
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        self.check_for_root()
        method = self.root_chain_manager.method('depositEtherFor', user_address)
        return self.process_write(method, option, private_key)

    def withdraw_start(
        self,
        amount: int,
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Initiate withdraw by burning provided amount."""
        self.check_for_child()
        method = self.method('withdraw', amount)
        return self.process_write(method, option, private_key)

    def transfer(
        self,
        amount: int,
        to: str,
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Transfer specified amount to another user."""
        return self.transfer_erc_20(to, amount, private_key, option)



from __future__ import annotations

from typing import Sequence

from matic.constants import LogEventSignature
from matic.json_types import IExitTransactionOption, ITransactionOption
from matic.pos.pos_token import TokenWithApproveAll


class ERC721(TokenWithApproveAll):
    """Arbitrary ERC-721-compliant token."""

    CONTRACT_NAME: str = 'ChildERC721'
    BURN_EVENT_SIGNATURE: bytes = LogEventSignature.ERC_721_TRANSFER

    @staticmethod
    def _validate_many(token_ids: Sequence[int]) -> list[int]:
        """Assert sequence length is less than 20."""
        if len(token_ids) > 20:
            raise ValueError('can not process more than 20 tokens')

        return list(token_ids)

    def get_tokens_count(
        self, user_address: str, options: ITransactionOption | None = None
    ) -> int:
        """Get tokens count for the user."""
        method = self.method('balanceOf', user_address)
        return int(self.process_read(method, options))

    def get_token_id_at_index_for_user(
        self, index: int, user_address: str, options: ITransactionOption | None = None
    ) -> int:
        """Get token id on supplied index for user."""
        method = self.method('tokenOfOwnerByIndex', user_address, index)

        return int(self.process_read(method, options))

    def get_all_tokens(self, user_address: str, limit: int | None = None) -> list[int]:
        """Get all tokens for user."""
        count = self.get_tokens_count(user_address)
        if limit is not None and count > limit:
            count = limit

        # TODO: should be async
        return [
            self.get_token_id_at_index_for_user(i, user_address) for i in range(count)
        ]

    def is_approved(
        self, token_id: int, option: ITransactionOption | None = None
    ) -> bool:
        """Check if given token is approved for contract."""
        self.check_for_root()
        method = self.method('getApproved', token_id)
        return self.predicate_address == self.process_read(method, option)

    def approve(
        self,
        token_id: int,
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Approve token with given id to contract (root chain)."""
        self.check_for_root()
        method = self.method('approve', self.predicate_address, token_id)
        return self.process_write(method, option, private_key)

    def deposit(
        self,
        token_id: int,
        user_address: str,
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Deposit given token from root chain to child."""
        self.check_for_root()
        amount_in_abi = self.client.parent.encode_parameters([token_id], ['uint256'])
        return self.root_chain_manager.deposit(
            user_address,
            self.address,
            amount_in_abi,
            private_key,
            option,
        )

    def deposit_many(
        self,
        token_ids: Sequence[int],
        user_address: str,
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Deposit given tokens from root chain to child."""
        self.check_for_root()
        self._validate_many(token_ids)
        amount_in_abi = self.client.parent.encode_parameters([token_ids], ['uint256[]'])
        return self.root_chain_manager.deposit(
            user_address,
            self.address,
            amount_in_abi,
            private_key,
            option,
        )

    def withdraw_start(
        self,
        token_id: int,
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Begin withdrawal to root chain (on child chain)."""
        self.check_for_child()
        method = self.method('withdraw', token_id)
        return self.process_write(method, option, private_key)

    def withdraw_start_with_metadata(
        self,
        token_id: int,
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Begin withdrawal to root chain with writing of metadata (on child chain)."""
        self.check_for_child()
        method = self.method('withdrawWithMetadata', token_id)
        return self.process_write(method, option, private_key)

    def withdraw_start_many(
        self,
        token_ids: Sequence[int],
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Begin withdrawal of multiple tokens to root chain (on child chain)."""
        self.check_for_child()
        self._validate_many(token_ids)
        method = self.method('withdrawBatch', token_ids)
        return self.process_write(method, option, private_key)

    def withdraw_exit_on_index(
        self,
        burn_transaction_hash: bytes,
        index: int,
        private_key: str | None = None,
        option: IExitTransactionOption | None = None,
    ):
        """Complete withdraw process for token on given index.

        This function should be called after checkpoint has been submitted
        for the block containing burn tx.

        This function uses API to fetch proof data.
        """
        return self.withdraw_exit_pos(
            burn_transaction_hash,
            self.BURN_EVENT_SIGNATURE,
            private_key,
            True,
            index,
            option=option,
        )

    def is_withdraw_exited_many(self, tx_hash: bytes) -> bool:
        """Check if batch withdrawal is already exited for given transaction."""
        return self.is_withdrawn(tx_hash, LogEventSignature.ERC_721_BATCH_TRANSFER)

    def is_withdraw_exited_on_index(self, tx_hash: bytes, index: int) -> bool:
        """Check if withdrawal is already exited for given transaction on index."""
        return self.is_withdrawn_on_index(
            tx_hash, index, LogEventSignature.ERC_721_TRANSFER
        )

    def transfer(
        self,
        token_id: int,
        from_: str,
        to: str,
        private_key: str | None = None,
        option: ITransactionOption | None = None,
    ):
        """Transfer to another user."""
        return self.transfer_erc_721(from_, to, token_id, private_key, option)
